
from django.conf import settings
from django.db import models
from django.contrib.auth.models import Permission

class MissingPermission(models.Model):
    user = models.ForeignKey(
        settings.AUTH_USER_MODEL, on_delete=models.CASCADE, related_name="missing_perms"
    )
    permission = models.ForeignKey(
        Permission, on_delete=models.CASCADE, related_name="missing_by_users"
    )

    hits = models.PositiveIntegerField(default=1)
    first_seen = models.DateTimeField(auto_now_add=True)
    last_seen = models.DateTimeField(auto_now=True)

    # Nota deduplicata (IP/URL/UA, ecc.)
    note = models.TextField(blank=True, default="")

    class Meta:
        constraints = [
            models.UniqueConstraint(
                fields=["user", "permission"], name="uniq_user_permission"
            )
        ]
        ordering = ["-last_seen"]

    def __str__(self):
        return f"{self.user} missing {self.permission.content_type.app_label}.{self.permission.codename}"



def get_urlpatterns(base_url=None):
    from django.urls import include, path

    from django_auto_drf.auto_register import api_router_urls
    from django_auto_drf.settings import DJANGO_AUTO_DRF_DEFAULT_BASE_URL, DJANGO_AUTO_DRF_DISABLE_SCHEMA_GENERATION
    baseurl = base_url or DJANGO_AUTO_DRF_DEFAULT_BASE_URL
    if not baseurl.endswith("/"):
        baseurl += "/"
    if baseurl.startswith("/"):
        baseurl = baseurl[1:]
    urlpatterns = [
        path(baseurl, include(api_router_urls())),
    ]

    if not DJANGO_AUTO_DRF_DISABLE_SCHEMA_GENERATION:
        from drf_spectacular.views import SpectacularAPIView, SpectacularRedocView, SpectacularSwaggerView
        from drf_spectacular.renderers import OpenApiJsonRenderer

        urlpatterns += [
            path("api/schema/", SpectacularAPIView.as_view(), name="schema"),
            path("api/schema.json", SpectacularAPIView.as_view(renderer_classes=[OpenApiJsonRenderer]), name="schema-json"),
            path("api/docs/", SpectacularSwaggerView.as_view(url_name="schema"), name="swagger-ui"),
            path("api/redoc/", SpectacularRedocView.as_view(url_name="schema"), name="redoc"),
        ]
    return urlpatterns



from rest_framework.views import exception_handler as drf_exception_handler
from .utils import (
    guess_model_from_view,
    required_perm_strings,
    resolve_permissions,
    record_missing_perm,
    build_snippet,
)


def audit_exception_handler(exc, context):
    response = drf_exception_handler(exc, context)
    if response is None or response.status_code != 403:
        return response

    request = context.get("request")
    view = context.get("view")
    user = getattr(request, "user", None)
    if not (user and user.is_authenticated):
        return response  # ignore anonymous

    app_model = guess_model_from_view(view)
    if not app_model:
        return response

    app_label, model_name = app_model
    perm_strings = required_perm_strings(app_label, model_name, request.method)
    permissions = resolve_permissions(perm_strings)
    if not permissions:
        return response

    snippet = build_snippet(
        ip=request.META.get("REMOTE_ADDR", ""),
        url=request.build_absolute_uri() if hasattr(request, "build_absolute_uri") else request.path,
        ua=request.META.get("HTTP_USER_AGENT", ""),
        method=getattr(request, "method", ""),
    )

    for perm in permissions:
        record_missing_perm(user, perm, snippet=snippet)

    return response



from typing import Optional, Tuple, List
from django.contrib.auth.models import Permission
from django.db import transaction

from .models import MissingPermission

# --- Permission probing utilities (from perm_probe.py) ---
DEFAULT_PERMS_MAP = {
    "GET":    ["%(app_label)s.view_%(model_name)s"],
    "OPTIONS":["%(app_label)s.view_%(model_name)s"],
    "HEAD":   ["%(app_label)s.view_%(model_name)s"],
    "POST":   ["%(app_label)s.add_%(model_name)s"],
    "PUT":    ["%(app_label)s.change_%(model_name)s"],
    "PATCH":  ["%(app_label)s.change_%(model_name)s"],
    "DELETE": ["%(app_label)s.delete_%(model_name)s"],
}


def guess_model_from_view(view) -> Optional[Tuple[str, str]]:
    model = None
    if hasattr(view, "get_queryset"):
        try:
            qs = view.get_queryset()
            if qs is not None and hasattr(qs, "model"):
                model = qs.model
        except Exception:
            pass
    if model is None and hasattr(view, "serializer_class"):
        ser = getattr(view, "serializer_class", None)
        model = getattr(getattr(ser, "Meta", None), "model", None) or model
    if model is None:
        return None
    return model._meta.app_label, model._meta.model_name


def required_perm_strings(app_label: str, model_name: str, method: str) -> List[str]:
    tpl = DEFAULT_PERMS_MAP.get(method.upper(), [])
    return [t % {"app_label": app_label, "model_name": model_name} for t in tpl]


def resolve_permissions(perm_strings: List[str]) -> List[Permission]:
    result = []
    for s in perm_strings:
        try:
            app_label, codename = s.split(".", 1)
            p = Permission.objects.select_related("content_type").get(
                content_type__app_label=app_label, codename=codename
            )
            result.append(p)
        except Exception:
            continue
    return result


# --- Recording utilities (from record.py) ---

def _dedup_append(existing: str, snippet: str, sep: str = "\n") -> str:
    snippet = (snippet or "").strip()
    if not snippet:
        return existing or ""
    existing = existing or ""
    if snippet in existing:
        return existing
    return f"{existing}{sep}{snippet}" if existing else snippet


def build_snippet(ip: str = "", url: str = "", ua: str = "", method: str = "") -> str:
    parts = []
    if ip:
        parts.append(f"ip={ip}")
    if method:
        parts.append(f"method={method}")
    if url:
        parts.append(f"url={url}")
    if ua:
        parts.append(f"ua={ua[:256]}")
    return " | ".join(parts)


@transaction.atomic
def record_missing_perm(user, permission, *, snippet: str = ""):
    obj, created = MissingPermission.objects.select_for_update().get_or_create(
        user=user,
        permission=permission,
        defaults={"note": snippet or ""},
    )
    if not created:
        obj.hits += 1
        obj.note = _dedup_append(obj.note, snippet)
    obj.save()
    return obj



import logging

from rest_framework.exceptions import (
    AuthenticationFailed, NotAuthenticated, PermissionDenied
)
from rest_framework.views import exception_handler as drf_exception_handler

logger = logging.getLogger(__name__)


def drf_exception_handler_logger(exc, context):
    request = context.get("request")
    user = getattr(request, 'user', None)
    ip = get_client_ip(request)
    method = request.method
    path = request.get_full_path()

    # Log di accessi non autorizzati
    if isinstance(exc, (AuthenticationFailed, NotAuthenticated, PermissionDenied)):
        if user and user.is_authenticated:
            user_info = f"user={user.username} (id={user.id})"
        else:
            user_info = "utente non autenticato"

        logger.warning(
            f"[{exc.__class__.__name__}] {user_info} - {method} {path} da IP={ip} - payload={safe_data(request.data)}"
        )

    return drf_exception_handler(exc, context)


def get_client_ip(request):
    x_forwarded_for = request.META.get("HTTP_X_FORWARDED_FOR")
    if x_forwarded_for:
        return x_forwarded_for.split(",")[0]
    return request.META.get("REMOTE_ADDR")


def safe_data(data):
    if not hasattr(data, 'copy'):
        return {}
    data = data.copy()
    for key in ['password', 'token', 'access', 'refresh']:
        if key in data:
            data[key] = '***'
    return dict(data)



from rest_framework.permissions import DjangoModelPermissions


class StrictDjangoModelPermissions(DjangoModelPermissions):
    perms_map = DjangoModelPermissions.perms_map.copy()
    perms_map.update({
        'GET': ['%(app_label)s.view_%(model_name)s'],
        'HEAD': ['%(app_label)s.view_%(model_name)s'],
        'OPTIONS': ['%(app_label)s.view_%(model_name)s'],
    })




from django.contrib import admin, messages
from django.contrib.auth.models import Group
from django.urls import path, reverse
from django.shortcuts import render, redirect
from django import forms
from django.db import transaction

from .models import MissingPermission

class GroupSelectForm(forms.Form):
    group = forms.ModelChoiceField(
        queryset=Group.objects.all(),
        required=True,
        label="Gruppo di destinazione",
        help_text="Il permesso negato verrà aggiunto a questo gruppo."
    )

@admin.register(MissingPermission)
class MissingPermissionAdmin(admin.ModelAdmin):
    list_display = ("user", "permission", "hits", "last_seen")
    list_filter = ("permission__content_type__app_label", "permission__codename")
    search_fields = ("user__username", "user__email", "permission__codename", "note")
    readonly_fields = ("user", "permission", "hits", "first_seen", "last_seen", "note")

    actions = [
        "grant_permission_to_user",
        "clear_notes",
        "assign_permissions_to_group",
    ]

    def has_add_permission(self, request): return False
    def has_change_permission(self, request, obj=None): return False

    @admin.action(description="Concedi questo permesso all’utente (ATTENZIONE)")
    def grant_permission_to_user(self, request, queryset):
        updated = 0
        with transaction.atomic():
            for rec in queryset:
                rec.user.user_permissions.add(rec.permission)
                updated += 1
        if updated:
            self.message_user(request, f"Assegnati {updated} permessi agli utenti selezionati.", level=messages.SUCCESS)
        else:
            self.message_user(request, "Nessun permesso assegnato.", level=messages.WARNING)

    @admin.action(description="Pulisci la nota")
    def clear_notes(self, request, queryset):
        queryset.update(note="")
        self.message_user(request, "Note pulite.", level=messages.INFO)

    @admin.action(description="Aggiungi questo permesso a un Gruppo…")
    def assign_permissions_to_group(self, request, queryset):
        ids = ",".join(str(pk) for pk in queryset.values_list("pk", flat=True))
        url = f"{reverse('admin:audit_missingpermission_assign_to_group')}?ids={ids}"
        return redirect(url)

    def get_urls(self):
        urls = super().get_urls()
        custom = [
            path(
                "assign-to-group/",
                self.admin_site.admin_view(self.assign_to_group_view),
                name="audit_missingpermission_assign_to_group",
            ),
        ]
        return custom + urls

    def assign_to_group_view(self, request):
        ids_param = request.GET.get("ids") if request.method == "GET" else request.POST.get("ids")
        if not ids_param:
            self.message_user(request, "Nessun elemento selezionato.", level=messages.WARNING)
            return redirect("admin:audit_missingpermission_changelist")

        try:
            ids = [int(x) for x in ids_param.split(",") if x.strip()]
        except ValueError:
            self.message_user(request, "Lista ID non valida.", level=messages.ERROR)
            return redirect("admin:audit_missingpermission_changelist")

        qs = MissingPermission.objects.filter(pk__in=ids)

        if request.method == "POST":
            form = GroupSelectForm(request.POST)
            if form.is_valid():
                group = form.cleaned_data["group"]
                with transaction.atomic():
                    count_added = 0
                    for rec in qs.select_related("permission"):
                        group.permissions.add(rec.permission)
                        count_added += 1
                self.message_user(
                    request,
                    f"Assegnati {count_added} permessi al gruppo “{group.name}”.",
                    level=messages.SUCCESS,
                )
                return redirect("admin:audit_missingpermission_changelist")
        else:
            form = GroupSelectForm()

        context = {
            **self.admin_site.each_context(request),
            "title": "Aggiungi permessi selezionati a un Gruppo",
            "form": form,
            "ids": ids_param,
            "queryset": qs,
        }
        return render(request, "admin/audit_assign_to_group.html", context)



# Generated by Django 5.1.7 on 2025-08-26 11:49

import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('auth', '0012_alter_user_first_name_max_length'),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='MissingPermission',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('hits', models.PositiveIntegerField(default=1)),
                ('first_seen', models.DateTimeField(auto_now_add=True)),
                ('last_seen', models.DateTimeField(auto_now=True)),
                ('note', models.TextField(blank=True, default='')),
                ('permission', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='missing_by_users', to='auth.permission')),
                ('user', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='missing_perms', to=settings.AUTH_USER_MODEL)),
            ],
            options={
                'ordering': ['-last_seen'],
                'constraints': [models.UniqueConstraint(fields=('user', 'permission'), name='uniq_user_permission')],
            },
        ),
    ]



from django.apps import AppConfig


class AuditConfig(AppConfig):
    default_auto_field = "django.db.models.BigAutoField"
    name = "django_auto_drf.audit"
    verbose_name = "Audit – Missing Permissions"
